@@ -30,13 +30,16 @@
 #include "gimp-edit.h"
 #include "gimpbuffer.h"
 #include "gimpcontext.h"
+#include "gimpdrawable-edit.h"
 #include "gimpgrouplayer.h"
 #include "gimpimage.h"
 #include "gimpimage-duplicate.h"
+#include "gimpimage-merge.h"
 #include "gimpimage-new.h"
 #include "gimpimage-undo.h"
 #include "gimplayer-floating-selection.h"
 #include "gimplayer-new.h"
+#include "gimplayermask.h"
 #include "gimplist.h"
 #include "gimppickable.h"
 #include "gimpselection.h"
@@ -46,11 +49,12 @@
 
 /*  local function protypes  */
 
-static GimpBuffer * gimp_edit_extract (GimpImage     *image,
-                                       GList         *pickables,
-                                       GimpContext   *context,
-                                       gboolean       cut_pixels,
-                                       GError       **error);
+static GimpBuffer   * gimp_edit_extract            (GimpImage     *image,
+                                                    GList         *pickables,
+                                                    GimpContext   *context,
+                                                    gboolean       cut_pixels,
+                                                    GError       **error);
+static GimpDrawable * gimp_edit_paste_get_top_item (GList         *drawables);
 
 
 /*  public functions  */
@@ -115,7 +119,7 @@ gimp_edit_cut (GimpImage     *image,
       g_list_free (remove);
 
       /* Now copy all layers into the clipboard image. */
-      clip_image = gimp_image_new_from_drawables (image->gimp, drawables, FALSE);
+      clip_image = gimp_image_new_from_drawables (image->gimp, drawables, FALSE, TRUE);
       gimp_container_remove (image->gimp->images, GIMP_OBJECT (clip_image));
       gimp_set_clipboard_image (image->gimp, clip_image);
       g_object_unref (clip_image);
@@ -181,38 +185,64 @@ gimp_edit_copy (GimpImage     *image,
   /* Only accept multiple drawables for layers. */
   g_return_val_if_fail (g_list_length (drawables) == 1 || drawables_are_layers, NULL);
 
-  if (drawables_are_layers &&
-      gimp_channel_is_empty (gimp_image_get_mask (image)))
+  if (drawables_are_layers)
     {
       /* Special-casing the 1 layer with no selection case.
        * It allows us to save the whole layer with all pixels as stored,
        * not the rendered version of it.
        */
-      GimpImage *clip_image;
+      GimpImage   *clip_image;
+      GimpChannel *clip_selection;
 
-      clip_image = gimp_image_new_from_drawables (image->gimp, drawables, FALSE);
+      clip_image = gimp_image_new_from_drawables (image->gimp, drawables, TRUE, TRUE);
       gimp_container_remove (image->gimp->images, GIMP_OBJECT (clip_image));
       gimp_set_clipboard_image (image->gimp, clip_image);
       g_object_unref (clip_image);
 
+      clip_selection = gimp_image_get_mask (clip_image);
+      if (! gimp_channel_is_empty (clip_selection))
+        {
+          GList         *all_items;
+          GeglRectangle  selection_bounds;
+
+          gimp_item_bounds (GIMP_ITEM (clip_selection),
+                            &selection_bounds.x, &selection_bounds.y,
+                            &selection_bounds.width, &selection_bounds.height);
+
+          /* Invert the selection. */
+          gimp_channel_invert (clip_selection, FALSE);
+          all_items = gimp_image_get_layer_list (clip_image);
+
+          for (iter = all_items; iter; iter = g_list_next (iter))
+            {
+              gint item_x;
+              gint item_y;
+
+              gimp_item_get_offset (GIMP_ITEM (iter->data), &item_x, &item_y);
+
+              /* Even if the original layer may not have an alpha channel, the
+               * selected data must always have one. First because a selection
+               * is in some way an alpha channel when we copy (we may copy part
+               * of a pixel, i.e. with transparency). Second because the
+               * selection is not necessary rectangular, unlike layers. So when
+               * we will clear, if we hadn't added an alpha channel, we'd end up
+               * with background color all over the place.
+               */
+              gimp_layer_add_alpha (GIMP_LAYER (iter->data));
+              gimp_drawable_edit_clear (GIMP_DRAWABLE (iter->data), context);
+
+              /* Finally shrink the copied layer to selection bounds. */
+              gimp_item_resize (iter->data, context, GIMP_FILL_TRANSPARENT,
+                                selection_bounds.width, selection_bounds.height,
+                                item_x - selection_bounds.x, item_y - selection_bounds.y);
+            }
+          g_list_free (all_items);
+        }
+      /* Remove selection from the clipboard image. */
+      gimp_channel_clear (clip_selection, NULL, FALSE);
+
       return GIMP_OBJECT (gimp_get_clipboard_image (image->gimp));
     }
-  else if (drawables_are_layers)
-    {
-      /* Copying multiple layers or specific selection, we copy the
-       * composited pixels as rendered.
-       */
-      GimpImage  *clip_image;
-      GimpBuffer *buffer;
-
-      clip_image = gimp_image_new_from_drawables (image->gimp, drawables, TRUE);
-      gimp_container_remove (image->gimp->images, GIMP_OBJECT (clip_image));
-      buffer = gimp_edit_copy_visible (clip_image, context, error);
-
-      g_object_unref (clip_image);
-
-      return buffer ? GIMP_OBJECT (buffer) : NULL;
-    }
   else
     {
       GimpBuffer *buffer;
@@ -266,11 +296,15 @@ gimp_edit_paste_is_in_place (GimpPasteType paste_type)
     case GIMP_PASTE_TYPE_FLOATING:
     case GIMP_PASTE_TYPE_FLOATING_INTO:
     case GIMP_PASTE_TYPE_NEW_LAYER:
+    case GIMP_PASTE_TYPE_NEW_LAYER_OR_FLOATING:
+    case GIMP_PASTE_TYPE_NEW_MERGED_LAYER_OR_FLOATING:
       return FALSE;
 
     case GIMP_PASTE_TYPE_FLOATING_IN_PLACE:
     case GIMP_PASTE_TYPE_FLOATING_INTO_IN_PLACE:
     case GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE:
+    case GIMP_PASTE_TYPE_NEW_LAYER_OR_FLOATING_IN_PLACE:
+    case GIMP_PASTE_TYPE_NEW_MERGED_LAYER_OR_FLOATING_IN_PLACE:
       return TRUE;
     }
 
@@ -278,7 +312,8 @@ gimp_edit_paste_is_in_place (GimpPasteType paste_type)
 }
 
 static gboolean
-gimp_edit_paste_is_floating (GimpPasteType paste_type)
+gimp_edit_paste_is_floating (GimpPasteType  paste_type,
+                             GimpDrawable  *drawable)
 {
   switch (paste_type)
     {
@@ -291,89 +326,73 @@ gimp_edit_paste_is_floating (GimpPasteType paste_type)
     case GIMP_PASTE_TYPE_NEW_LAYER:
     case GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE:
       return FALSE;
+
+    case GIMP_PASTE_TYPE_NEW_LAYER_OR_FLOATING:
+    case GIMP_PASTE_TYPE_NEW_LAYER_OR_FLOATING_IN_PLACE:
+    case GIMP_PASTE_TYPE_NEW_MERGED_LAYER_OR_FLOATING:
+    case GIMP_PASTE_TYPE_NEW_MERGED_LAYER_OR_FLOATING_IN_PLACE:
+      if (GIMP_IS_LAYER_MASK (drawable))
+        return TRUE;
+      else
+        return FALSE;
     }
 
   g_return_val_if_reached (FALSE);
 }
 
 static GList *
-gimp_edit_paste_get_layers (GimpImage     *image,
-                            GimpDrawable  *drawable,
-                            GimpObject    *paste,
-                            GimpPasteType *paste_type)
+gimp_edit_paste_get_tagged_layers (GimpImage         *image,
+                                   GList             *layers,
+                                   GList             *returned_layers,
+                                   const Babl        *floating_format,
+                                   GimpImageBaseType  base_type,
+                                   GimpPrecision      precision,
+                                   GimpPasteType      paste_type)
 {
-  GList      *layers = NULL;
-  const Babl *floating_format;
+  GList *iter;
 
-  /*  change paste type to NEW_LAYER for cases where we can't attach a
-   *  floating selection
-   */
-  if (! drawable                                            ||
-      gimp_viewable_get_children (GIMP_VIEWABLE (drawable)) ||
-      gimp_item_is_content_locked (GIMP_ITEM (drawable), NULL))
+  for (iter = layers; iter; iter = iter->next)
     {
-      if (gimp_edit_paste_is_in_place (*paste_type))
-        *paste_type = GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE;
-      else
-        *paste_type = GIMP_PASTE_TYPE_NEW_LAYER;
-    }
+      GimpLayer *layer;
+      GType      layer_type;
+      gboolean   copied = TRUE;
 
-  /*  floating pastes always have the pasted-to drawable's format with
-   *  alpha; if drawable == NULL, user is pasting into an empty image
-   */
-  if (drawable && gimp_edit_paste_is_floating (*paste_type))
-    floating_format = gimp_drawable_get_format_with_alpha (drawable);
-  else
-    floating_format = gimp_image_get_layer_format (image, TRUE);
-
-  if (GIMP_IS_IMAGE (paste))
-    {
-      GList *iter;
-
-      layers = g_list_copy (gimp_image_get_layer_iter (GIMP_IMAGE (paste)));
-
-      if (g_list_length (layers) > 1)
+      switch (paste_type)
         {
-          if (gimp_edit_paste_is_in_place (*paste_type))
-            *paste_type = GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE;
+        case GIMP_PASTE_TYPE_FLOATING:
+        case GIMP_PASTE_TYPE_FLOATING_IN_PLACE:
+        case GIMP_PASTE_TYPE_FLOATING_INTO:
+        case GIMP_PASTE_TYPE_FLOATING_INTO_IN_PLACE:
+          /*  when pasting as floating make sure gimp_item_convert()
+           *  will turn group layers into normal layers, otherwise use
+           *  the same layer type so e.g. text information gets
+           *  preserved. See issue #2667.
+           */
+          if (GIMP_IS_GROUP_LAYER (iter->data))
+            layer_type = GIMP_TYPE_LAYER;
           else
-            *paste_type = GIMP_PASTE_TYPE_NEW_LAYER;
+            layer_type = G_TYPE_FROM_INSTANCE (iter->data);
+          break;
+
+        case GIMP_PASTE_TYPE_NEW_LAYER:
+        case GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE:
+          layer_type = G_TYPE_FROM_INSTANCE (iter->data);
+          break;
+
+        default:
+          g_return_val_if_reached (NULL);
         }
 
-      for (iter = layers; iter; iter = iter->next)
+      if (GIMP_IS_GROUP_LAYER (iter->data))
+        copied = (gboolean) GPOINTER_TO_INT (g_object_get_data (G_OBJECT (iter->data),
+                                                                "gimp-image-copied-layer"));
+      if (copied)
         {
-          GType layer_type;
+          layer = GIMP_LAYER (gimp_item_convert (GIMP_ITEM (iter->data),
+                                                 image, layer_type));
+          returned_layers = g_list_prepend (returned_layers, layer);
 
-          switch (*paste_type)
-            {
-            case GIMP_PASTE_TYPE_FLOATING:
-            case GIMP_PASTE_TYPE_FLOATING_IN_PLACE:
-            case GIMP_PASTE_TYPE_FLOATING_INTO:
-            case GIMP_PASTE_TYPE_FLOATING_INTO_IN_PLACE:
-              /*  when pasting as floating make sure gimp_item_convert()
-               *  will turn group layers into normal layers, otherwise use
-               *  the same layer type so e.g. text information gets
-               *  preserved. See issue #2667.
-               */
-              if (GIMP_IS_GROUP_LAYER (iter->data))
-                layer_type = GIMP_TYPE_LAYER;
-              else
-                layer_type = G_TYPE_FROM_INSTANCE (iter->data);
-              break;
-
-            case GIMP_PASTE_TYPE_NEW_LAYER:
-            case GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE:
-              layer_type = G_TYPE_FROM_INSTANCE (iter->data);
-              break;
-
-            default:
-              g_return_val_if_reached (NULL);
-            }
-
-          iter->data = GIMP_LAYER (gimp_item_convert (GIMP_ITEM (iter->data),
-                                                      image, layer_type));
-
-          switch (*paste_type)
+          switch (paste_type)
             {
             case GIMP_PASTE_TYPE_FLOATING:
             case GIMP_PASTE_TYPE_FLOATING_IN_PLACE:
@@ -389,10 +408,8 @@ gimp_edit_paste_get_layers (GimpImage     *image,
                   floating_format)
                 {
                   gimp_drawable_convert_type (GIMP_DRAWABLE (iter->data), image,
-                                              gimp_drawable_get_base_type (drawable),
-                                              gimp_drawable_get_precision (drawable),
-                                              TRUE,
-                                              NULL, NULL,
+                                              base_type, precision,
+                                              TRUE, NULL, NULL,
                                               GEGL_DITHER_NONE, GEGL_DITHER_NONE,
                                               FALSE, NULL);
                 }
@@ -402,6 +419,69 @@ gimp_edit_paste_get_layers (GimpImage     *image,
               break;
             }
         }
+      else
+        {
+          GimpContainer *container;
+
+          container = gimp_viewable_get_children (iter->data);
+          returned_layers = gimp_edit_paste_get_tagged_layers (image,
+                                                               GIMP_LIST (container)->queue->head,
+                                                               returned_layers,
+                                                               floating_format,
+                                                               base_type, precision, paste_type);
+        }
+    }
+
+  return returned_layers;
+}
+
+static GList *
+gimp_edit_paste_get_layers (GimpImage     *image,
+                            GList         *drawables,
+                            GimpObject    *paste,
+                            GimpPasteType *paste_type)
+{
+  GList      *layers = NULL;
+  const Babl *floating_format;
+
+  /*  change paste type to NEW_LAYER for cases where we can't attach a
+   *  floating selection
+   */
+  if (g_list_length (drawables) != 1               ||
+      gimp_viewable_get_children (drawables->data) ||
+      gimp_item_is_content_locked (drawables->data, NULL))
+    {
+      if (gimp_edit_paste_is_in_place (*paste_type))
+        *paste_type = GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE;
+      else
+        *paste_type = GIMP_PASTE_TYPE_NEW_LAYER;
+    }
+
+  /*  floating pastes always have the pasted-to drawable's format with
+   *  alpha; if drawable == NULL, user is pasting into an empty image
+   */
+  if (drawables && gimp_edit_paste_is_floating (*paste_type, drawables->data))
+    floating_format = gimp_drawable_get_format_with_alpha (drawables->data);
+  else
+    floating_format = gimp_image_get_layer_format (image, TRUE);
+
+  if (GIMP_IS_IMAGE (paste))
+    {
+      layers = gimp_image_get_layer_iter (GIMP_IMAGE (paste));
+
+      if (g_list_length (layers) > 1)
+        {
+          if (gimp_edit_paste_is_in_place (*paste_type))
+            *paste_type = GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE;
+          else
+            *paste_type = GIMP_PASTE_TYPE_NEW_LAYER;
+        }
+
+      layers = gimp_edit_paste_get_tagged_layers (image, layers, NULL, floating_format,
+                                                  gimp_drawable_get_base_type (drawables->data),
+                                                  gimp_drawable_get_precision (drawables->data),
+                                                  *paste_type);
+      layers = g_list_reverse (layers);
     }
   else if (GIMP_IS_BUFFER (paste))
     {
@@ -420,27 +500,32 @@ gimp_edit_paste_get_layers (GimpImage     *image,
 }
 
 static void
-gimp_edit_paste_get_viewport_offset (GimpImage    *image,
-                                     GimpDrawable *drawable,
-                                     GList        *pasted_layers,
-                                     gint          viewport_x,
-                                     gint          viewport_y,
-                                     gint          viewport_width,
-                                     gint          viewport_height,
-                                     gint         *offset_x,
-                                     gint         *offset_y)
+gimp_edit_paste_get_viewport_offset (GimpImage *image,
+                                     GList     *drawables,
+                                     GList     *pasted_layers,
+                                     gint       viewport_x,
+                                     gint       viewport_y,
+                                     gint       viewport_width,
+                                     gint       viewport_height,
+                                     gint      *pasted_bbox_x,
+                                     gint      *pasted_bbox_y,
+                                     gint      *offset_x,
+                                     gint      *offset_y)
 {
   GList   *iter;
   gint     image_width;
   gint     image_height;
-  gint     width          = 0;
-  gint     height         = 0;
+  /* Source: pasted layers */
+  gint     source_width;
+  gint     source_height;
+
+  gint     x1             = G_MAXINT;
+  gint     y1             = G_MAXINT;
+  gint     x2             = G_MININT;
+  gint     y2             = G_MININT;
   gboolean clamp_to_image = TRUE;
 
   g_return_if_fail (GIMP_IS_IMAGE (image));
-  g_return_if_fail (drawable == NULL || GIMP_IS_DRAWABLE (drawable));
-  g_return_if_fail (drawable == NULL ||
-                    gimp_item_is_attached (GIMP_ITEM (drawable)));
   g_return_if_fail (pasted_layers != NULL);
   g_return_if_fail (offset_x != NULL);
   g_return_if_fail (offset_y != NULL);
@@ -450,16 +535,30 @@ gimp_edit_paste_get_viewport_offset (GimpImage    *image,
 
   for (iter = pasted_layers; iter; iter = iter->next)
     {
+      gint layer_off_x;
+      gint layer_off_y;
       gint layer_width;
       gint layer_height;
 
       g_return_if_fail (GIMP_IS_VIEWABLE (iter->data));
+
+      gimp_item_get_offset (GIMP_ITEM (iter->data), &layer_off_x, &layer_off_y);
       gimp_viewable_get_size (iter->data, &layer_width, &layer_height);
 
-      width  = MAX (width, layer_width);
-      height = MAX (height, layer_height);
+      x1 = MIN (layer_off_x, x1);
+      y1 = MIN (layer_off_y, y1);
+      x2 = MAX (layer_off_x + layer_width, x2);
+      y2 = MAX (layer_off_y + layer_height, y2);
     }
 
+  /* Source offset and dimensions: this is the bounding box for all layers which
+   * we want to paste together, keeping their relative position to each others.
+   */
+  *pasted_bbox_x = x1;
+  *pasted_bbox_y = y1;
+  source_width = x2 - x1;
+  source_height = y2 - y1;
+
   if (viewport_width  == image_width &&
       viewport_height == image_height)
     {
@@ -471,25 +570,34 @@ gimp_edit_paste_get_viewport_offset (GimpImage    *image,
       viewport_height = 0;
     }
 
-  if (drawable)
+  if (drawables)
     {
-      /*  if pasting to a drawable  */
-
-      GimpContainer *children;
-      gint           off_x, off_y;
-      gint           target_x, target_y;
-      gint           target_width, target_height;
-      gint           paste_x, paste_y;
-      gint           paste_width, paste_height;
-      gboolean       have_mask;
+      /*  if pasting while 1 or more drawables are selected  */
+      gboolean empty_target = TRUE;
+      gint     target_x;
+      gint     target_y;
+      gint     target_width;
+      gint     target_height;
+      gint     paste_x, paste_y;
+      gint     paste_width;
+      gint     paste_height;
+      gboolean have_mask;
 
       have_mask = ! gimp_channel_is_empty (gimp_image_get_mask (image));
 
-      gimp_item_get_offset (GIMP_ITEM (drawable), &off_x, &off_y);
+      for (iter = drawables; iter; iter = iter->next)
+        {
+          GimpContainer *children;
 
-      children = gimp_viewable_get_children (GIMP_VIEWABLE (drawable));
+          children = gimp_viewable_get_children (iter->data);
+          if (! children || gimp_container_get_n_children (children) > 0)
+            {
+              empty_target = FALSE;
+              break;
+            }
+        }
 
-      if (children && gimp_container_get_n_children (children) == 0)
+      if (empty_target)
         {
           /* treat empty layer groups as image-sized, use the selection
            * as target
@@ -500,36 +608,62 @@ gimp_edit_paste_get_viewport_offset (GimpImage    *image,
         }
       else
         {
-          gimp_item_mask_intersect (GIMP_ITEM (drawable),
-                                    &target_x, &target_y,
-                                    &target_width, &target_height);
+          gint x1 = G_MAXINT;
+          gint y1 = G_MAXINT;
+          gint x2 = G_MININT;
+          gint y2 = G_MININT;
+
+          for (iter = drawables; iter; iter = iter->next)
+            {
+              gint layer_off_x;
+              gint layer_off_y;
+              gint mask_off_x;
+              gint mask_off_y;
+              gint mask_width;
+              gint mask_height;
+
+              gimp_item_get_offset (GIMP_ITEM (iter->data), &layer_off_x, &layer_off_y);
+              /* This is the bounds relatively to the drawable. */
+              gimp_item_mask_intersect (GIMP_ITEM (iter->data),
+                                        &mask_off_x, &mask_off_y,
+                                        &mask_width, &mask_height);
+
+              /* The bounds relatively to the image. */
+              x1 = MIN (layer_off_x + mask_off_x, x1);
+              y1 = MIN (layer_off_y + mask_off_y, y1);
+              x2 = MAX (layer_off_x + mask_off_x + mask_width, x2);
+              y2 = MAX (layer_off_y + mask_off_y + mask_height, y2);
+            }
+
+          target_x      = x1;
+          target_y      = y1;
+          target_width  = x2 - x1;
+          target_height = y2 - y1;
         }
 
       if (! have_mask         &&    /* if we have no mask */
           viewport_width  > 0 &&    /* and we have a viewport */
           viewport_height > 0 &&
-          (width  < target_width || /* and the paste is smaller than the target */
-           height < target_height) &&
+          (source_width  < target_width || /* and the paste is smaller than the target */
+           source_height < target_height) &&
 
           /* and the viewport intersects with the target */
           gimp_rectangle_intersect (viewport_x, viewport_y,
                                     viewport_width, viewport_height,
-                                    off_x, off_y, /* target_x,y are 0 */
+                                    target_x, target_x,
                                     target_width, target_height,
                                     &paste_x, &paste_y,
                                     &paste_width, &paste_height))
         {
           /*  center on the viewport  */
-
-          *offset_x = paste_x + (paste_width - width)  / 2;
-          *offset_y = paste_y + (paste_height- height) / 2;
+          *offset_x = viewport_x + (viewport_width - source_width)  / 2;
+          *offset_y = viewport_y + (viewport_height- source_height) / 2;
         }
       else
         {
           /*  otherwise center on the target  */
-
-          *offset_x = off_x + target_x + (target_width  - width)  / 2;
-          *offset_y = off_y + target_y + (target_height - height) / 2;
+          *offset_x = target_x + (target_width  - source_width)  / 2;
+          *offset_y = target_y + (target_height - source_height) / 2;
 
           /*  and keep it that way  */
           clamp_to_image = FALSE;
@@ -537,20 +671,20 @@ gimp_edit_paste_get_viewport_offset (GimpImage    *image,
     }
   else if (viewport_width  > 0 &&  /* if we have a viewport */
            viewport_height > 0 &&
-           (width  < image_width || /* and the paste is       */
-            height < image_height)) /* smaller than the image */
+           (source_width  < image_width || /* and the paste is       */
+            source_height < image_height)) /* smaller than the image */
     {
       /*  center on the viewport  */
 
-      *offset_x = viewport_x + (viewport_width  - width)  / 2;
-      *offset_y = viewport_y + (viewport_height - height) / 2;
+      *offset_x = viewport_x + (viewport_width  - source_width)  / 2;
+      *offset_y = viewport_y + (viewport_height - source_height) / 2;
     }
   else
     {
       /*  otherwise center on the image  */
 
-      *offset_x = (image_width  - width)  / 2;
-      *offset_y = (image_height - height) / 2;
+      *offset_x = (image_width  - source_width)  / 2;
+      *offset_y = (image_height - source_height) / 2;
 
       /*  and keep it that way  */
       clamp_to_image = FALSE;
@@ -561,8 +695,8 @@ gimp_edit_paste_get_viewport_offset (GimpImage    *image,
       /*  Ensure that the pasted layer is always within the image, if it
        *  fits and aligned at top left if it doesn't. (See bug #142944).
        */
-      *offset_x = MIN (*offset_x, image_width  - width);
-      *offset_y = MIN (*offset_y, image_height - height);
+      *offset_x = MIN (*offset_x, image_width  - source_width);
+      *offset_y = MIN (*offset_y, image_height - source_height);
       *offset_x = MAX (*offset_x, 0);
       *offset_y = MAX (*offset_y, 0);
     }
@@ -570,15 +704,20 @@ gimp_edit_paste_get_viewport_offset (GimpImage    *image,
 
 static GList *
 gimp_edit_paste_paste (GimpImage     *image,
-                       GimpDrawable  *drawable,
+                       GList         *drawables,
                        GList         *layers,
                        GimpPasteType  paste_type,
                        gboolean       use_offset,
+                       gint           layers_bbox_x,
+                       gint           layers_bbox_y,
                        gint           offset_x,
                        gint           offset_y)
 {
   GList *iter;
 
+  g_return_val_if_fail (paste_type > GIMP_PASTE_TYPE_FLOATING_INTO_IN_PLACE ||
+                        g_list_length (drawables) == 1, NULL);
+
   gimp_image_undo_group_start (image, GIMP_UNDO_GROUP_EDIT_PASTE,
                                C_("undo-type", "Paste"));
 
@@ -586,7 +725,9 @@ gimp_edit_paste_paste (GimpImage     *image,
     {
       /* Layers for in-place paste are already translated. */
       if (use_offset)
-        gimp_item_translate (GIMP_ITEM (iter->data), offset_x, offset_y, FALSE);
+        gimp_item_translate (GIMP_ITEM (iter->data),
+                             offset_x - layers_bbox_x,
+                             offset_y - layers_bbox_y, FALSE);
 
       switch (paste_type)
         {
@@ -603,9 +744,21 @@ gimp_edit_paste_paste (GimpImage     *image,
 
         case GIMP_PASTE_TYPE_FLOATING_INTO:
         case GIMP_PASTE_TYPE_FLOATING_INTO_IN_PLACE:
-          floating_sel_attach (iter->data, drawable);
+          floating_sel_attach (iter->data, drawables->data);
           break;
 
+        case GIMP_PASTE_TYPE_NEW_LAYER_OR_FLOATING:
+        case GIMP_PASTE_TYPE_NEW_LAYER_OR_FLOATING_IN_PLACE:
+          if (g_list_length (drawables) == 1 && GIMP_IS_LAYER_MASK (drawables->data))
+            {
+              if (! gimp_channel_is_empty (gimp_image_get_mask (image)))
+                gimp_channel_clear (gimp_image_get_mask (image), NULL, TRUE);
+              floating_sel_attach (iter->data, drawables->data);
+              break;
+            }
+
+          /* fall thru if not a layer mask */
+
         case GIMP_PASTE_TYPE_NEW_LAYER:
         case GIMP_PASTE_TYPE_NEW_LAYER_IN_PLACE:
             {
@@ -615,15 +768,22 @@ gimp_edit_paste_paste (GimpImage     *image,
               /*  always add on top of a passed layer, where we would attach
                *  a floating selection
                */
-              if (GIMP_IS_LAYER (drawable))
+              if (g_list_length (drawables) > 0 && GIMP_IS_LAYER (drawables->data))
                 {
-                  parent   = gimp_layer_get_parent (GIMP_LAYER (drawable));
-                  position = gimp_item_get_index (GIMP_ITEM (drawable));
+                  GimpDrawable *top_drawable;
+
+                  top_drawable = gimp_edit_paste_get_top_item (drawables);
+                  parent   = gimp_layer_get_parent (GIMP_LAYER (top_drawable));
+                  position = gimp_item_get_index (GIMP_ITEM (top_drawable));
                 }
 
               gimp_image_add_layer (image, iter->data, parent, position, TRUE);
             }
           break;
+
+        case GIMP_PASTE_TYPE_NEW_MERGED_LAYER_OR_FLOATING:
+        case GIMP_PASTE_TYPE_NEW_MERGED_LAYER_OR_FLOATING_IN_PLACE:
+          g_return_val_if_reached (NULL);
         }
     }
 
@@ -634,26 +794,77 @@ gimp_edit_paste_paste (GimpImage     *image,
 
 GList *
 gimp_edit_paste (GimpImage     *image,
-                 GimpDrawable  *drawable,
+                 GList         *drawables,
                  GimpObject    *paste,
                  GimpPasteType  paste_type,
+                 GimpContext   *context,
+                 gboolean       merged,
                  gint           viewport_x,
                  gint           viewport_y,
                  gint           viewport_width,
                  gint           viewport_height)
 {
   GList    *layers;
-  gboolean  use_offset = FALSE;
-  gint      offset_x   = 0;
-  gint      offset_y   = 0;
+  gboolean  use_offset    = FALSE;
+  gint      layers_bbox_x = 0;
+  gint      layers_bbox_y = 0;
+  gint      offset_y      = 0;
+  gint      offset_x      = 0;
 
   g_return_val_if_fail (GIMP_IS_IMAGE (image), NULL);
-  g_return_val_if_fail (drawable == NULL || GIMP_IS_DRAWABLE (drawable), NULL);
-  g_return_val_if_fail (drawable == NULL ||
-                        gimp_item_is_attached (GIMP_ITEM (drawable)), NULL);
   g_return_val_if_fail (GIMP_IS_IMAGE (paste) || GIMP_IS_BUFFER (paste), NULL);
 
-  layers = gimp_edit_paste_get_layers (image, drawable, paste, &paste_type);
+  for (GList *iter = drawables; iter; iter = iter->next)
+    {
+      g_return_val_if_fail (GIMP_IS_DRAWABLE (iter->data), NULL);
+      g_return_val_if_fail (gimp_item_is_attached (GIMP_ITEM (iter->data)), NULL);
+    }
+
+  if (merged && GIMP_IS_IMAGE (paste))
+    {
+      GimpImage *tmp_image;
+
+      tmp_image = gimp_image_duplicate (GIMP_IMAGE (paste));
+      gimp_container_remove (image->gimp->images, GIMP_OBJECT (tmp_image));
+      gimp_image_merge_visible_layers (tmp_image, context, GIMP_EXPAND_AS_NECESSARY,
+                                       FALSE, FALSE, NULL);
+      layers = g_list_copy (gimp_image_get_layer_iter (tmp_image));
+
+      /* The merge process should ensure that we get a single non-group and
+       * no-mask layer.
+       */
+      g_return_val_if_fail (g_list_length (layers) == 1, NULL);
+
+      layers->data = gimp_item_convert (GIMP_ITEM (layers->data), image,
+                                        G_TYPE_FROM_INSTANCE (layers->data));
+
+      switch (paste_type)
+        {
+        case GIMP_PASTE_TYPE_FLOATING:
+        case GIMP_PASTE_TYPE_FLOATING_IN_PLACE:
+        case GIMP_PASTE_TYPE_FLOATING_INTO:
+        case GIMP_PASTE_TYPE_FLOATING_INTO_IN_PLACE:
+          if (gimp_drawable_get_format (GIMP_DRAWABLE (layers->data)) !=
+              gimp_drawable_get_format_with_alpha (GIMP_DRAWABLE (layers->data)))
+            {
+              gimp_drawable_convert_type (GIMP_DRAWABLE (layers->data), image,
+                                          gimp_drawable_get_base_type (layers->data),
+                                          gimp_drawable_get_precision (layers->data),
+                                          TRUE, NULL, NULL,
+                                          GEGL_DITHER_NONE, GEGL_DITHER_NONE,
+                                          FALSE, NULL);
+            }
+          break;
+
+        default:
+          break;
+        }
+      g_object_unref (tmp_image);
+    }
+  else
+    {
+      layers = gimp_edit_paste_get_layers (image, drawables, paste, &paste_type);
+    }
 
   if (! layers)
     return NULL;
@@ -672,17 +883,20 @@ gimp_edit_paste (GimpImage     *image,
   else
     {
       use_offset = TRUE;
-      gimp_edit_paste_get_viewport_offset (image, drawable, layers,
+      gimp_edit_paste_get_viewport_offset (image, drawables, layers,
                                            viewport_x,
                                            viewport_y,
                                            viewport_width,
                                            viewport_height,
+                                           &layers_bbox_x,
+                                           &layers_bbox_y,
                                            &offset_x,
                                            &offset_y);
     }
 
-  return gimp_edit_paste_paste (image, drawable, layers, paste_type,
-                                use_offset, offset_x, offset_y);
+  return gimp_edit_paste_paste (image, drawables, layers, paste_type,
+                                use_offset, layers_bbox_x, layers_bbox_y,
+                                offset_x, offset_y);
 }
 
 GimpImage *
@@ -873,3 +1087,54 @@ gimp_edit_extract (GimpImage     *image,
 
   return NULL;
 }
+
+/* Return the visually top item. */
+static GimpDrawable *
+gimp_edit_paste_get_top_item (GList *drawables)
+{
+  GList        *iter;
+  GimpDrawable *top      = NULL;
+  GList        *top_path = NULL;
+
+  for (iter = drawables; iter; iter = iter->next)
+    {
+      GList *path = gimp_item_get_path (iter->data);
+
+      if (top == NULL)
+        {
+          top      = iter->data;
+          top_path = path;
+          path     = NULL;
+        }
+      else
+        {
+          GList *p_iter;
+          GList *tp_iter;
+
+          for (p_iter = path, tp_iter = top_path; p_iter || tp_iter; p_iter = p_iter->next, tp_iter = tp_iter->next)
+            {
+              if (tp_iter == NULL)
+                {
+                  break;
+                }
+              else if (p_iter == NULL ||
+                       GPOINTER_TO_UINT (p_iter->data) < GPOINTER_TO_UINT (tp_iter->data))
+                {
+                  g_list_free (top_path);
+                  top_path = path;
+                  path = NULL;
+                  top = iter->data;
+                  break;
+                }
+              else if (GPOINTER_TO_UINT (p_iter->data) > GPOINTER_TO_UINT (tp_iter->data))
+                {
+                  break;
+                }
+            }
+        }
+      g_list_free (path);
+    }
+
+  g_list_free (top_path);
+  return top;
+}
