@@ -398,6 +398,86 @@ gimp_param_spec_layer (const gchar *name,
 }
 
 
+/*
+ * GIMP_TYPE_PARAM_TEXT_LAYER
+ */
+
+static void   gimp_param_text_layer_class_init (GParamSpecClass *klass);
+static void   gimp_param_text_layer_init       (GParamSpec      *pspec);
+
+GType
+gimp_param_text_layer_get_type (void)
+{
+  static GType type = 0;
+
+  if (! type)
+    {
+      const GTypeInfo info =
+      {
+        sizeof (GParamSpecClass),
+        NULL, NULL,
+        (GClassInitFunc) gimp_param_text_layer_class_init,
+        NULL, NULL,
+        sizeof (GimpParamSpecTextLayer),
+        0,
+        (GInstanceInitFunc) gimp_param_text_layer_init
+      };
+
+      type = g_type_register_static (GIMP_TYPE_PARAM_LAYER,
+                                     "GimpParamTextLayer", &info, 0);
+    }
+
+  return type;
+}
+
+static void
+gimp_param_text_layer_class_init (GParamSpecClass *klass)
+{
+  klass->value_type = GIMP_TYPE_TEXT_LAYER;
+}
+
+static void
+gimp_param_text_layer_init (GParamSpec *pspec)
+{
+}
+
+/**
+ * gimp_param_spec_text_layer:
+ * @name:    Canonical name of the property specified.
+ * @nick:    Nick name of the property specified.
+ * @blurb:   Description of the property specified.
+ * @none_ok: Whether no  is a valid value.
+ * @flags:   Flags for the property specified.
+ *
+ * Creates a new #GimpParamSpecTextLayer specifying a
+ * #GIMP_TYPE_TEXT_LAYER property.
+ *
+ * See g_param_spec_internal() for details on property names.
+ *
+ * Returns: (transfer full): The newly created #GimpParamSpecTextLayer.
+ *
+ * Since: 3.0
+ **/
+GParamSpec *
+gimp_param_spec_text_layer (const gchar *name,
+                       const gchar *nick,
+                       const gchar *blurb,
+                       gboolean     none_ok,
+                       GParamFlags  flags)
+{
+  GimpParamSpecItem *ispec;
+
+  ispec = g_param_spec_internal (GIMP_TYPE_PARAM_TEXT_LAYER,
+                                 name, nick, blurb, flags);
+
+  g_return_val_if_fail (ispec, NULL);
+
+  ispec->none_ok = none_ok ? TRUE : FALSE;
+
+  return G_PARAM_SPEC (ispec);
+}
+
+
 /*
  * GIMP_TYPE_PARAM_CHANNEL
  */
