@@ -169,7 +169,7 @@ package Gimp::CodeGen::pdb;
 		     init_value      => 'NULL',
 		     in_annotate     => '(element-type GimpChannel)',
 		     out_annotate    => '(element-type GimpChannel) (transfer container)',
-		     get_value_func  => '$var = gimp_value_get_object_array ($value)',
+		     get_value_func  => '$var = (const GimpChannel **) gimp_value_get_object_array ($value)',
 		     dup_value_func  => '{ GimpObjectArray *a = g_value_get_boxed (gimp_value_array_index ($value)); if (a) $var = g_memdup2 (a->data, a->length * sizeof (gpointer)); }',
 		     set_value_func  => 'gimp_value_set_object_array ($value, GIMP_TYPE_CHANNEL, (GObject **) $var, $var_len)',
 		     take_value_func => 'gimp_value_take_object_array ($value, GIMP_TYPE_CHANNEL, (GObject **) $var, $var_len)' },
@@ -248,6 +248,18 @@ package Gimp::CodeGen::pdb;
 		     take_value_func => 'g_value_set_object ($value, $var)',
 		     headers         => [ qw("core/gimplayer.h") ] },
 
+    text_layer  => { name            => 'TEXT_LAYER',
+		     gtype           => 'GIMP_TYPE_TEXT_LAYER',
+		     type            => 'GimpTextLayer *',
+		     const_type      => 'GimpTextLayer *',
+		     init_value      => 'NULL',
+		     out_annotate    => '(transfer none)',
+		     get_value_func  => '$var = g_value_get_object ($value)',
+		     dup_value_func  => '$var = GIMP_VALUES_GET_TEXT_LAYER ($value)',
+		     set_value_func  => 'g_value_set_object ($value, $var)',
+		     take_value_func => 'g_value_set_object ($value, $var)',
+		     headers         => [ qw("text/gimptextlayer.h") ] },
+
     channel     => { name            => 'CHANNEL',
 		     gtype           => 'GIMP_TYPE_CHANNEL',
 		     type            => 'GimpChannel *',
