@@ -815,7 +815,7 @@ mng_save_image (GFile         *file,
                 "jpeg-quality",    &config_jpeg_quality,
                 "jpeg-smoothing",  &config_jpeg_smoothing,
                 "loop",            &config_loop,
-                "defaut-delay",    &config_default_delay,
+                "default-delay",   &config_default_delay,
                 "default-chunks",  &config_default_chunks,
                 "default-dispose", &config_default_dispose,
                 "bkgd",            &config_bkgd,
