@@ -133,6 +133,7 @@ static void          reshuffle_cmap_write (guchar         *mapGimp);
 
 static void          save_header          (GOutputStream  *output,
                                            GimpImage      *image,
+                                           gboolean        export_cmyk,
                                            gboolean        export_duotone);
 
 static void          save_color_mode_data (GOutputStream  *output,
@@ -141,13 +142,22 @@ static void          save_color_mode_data (GOutputStream  *output,
 
 static void          save_resources       (GOutputStream  *output,
                                            GimpImage      *image,
+                                           gboolean        export_cmyk,
                                            gboolean        export_duotone);
 
-static void          save_layer_and_mask  (GOutputStream  *output,
+static void          save_paths           (GOutputStream  *output,
                                            GimpImage      *image);
 
+static void          save_layer_and_mask  (GOutputStream  *output,
+                                           GimpImage      *image,
+                                           gboolean        export_cmyk);
+
 static void          save_data            (GOutputStream  *output,
-                                           GimpImage      *image);
+                                           GimpImage      *image,
+                                           gboolean        export_cmyk);
+
+static void          double_to_psd_fixed  (gdouble         value,
+                                           gchar          *target);
 
 static void          xfwrite              (GOutputStream  *output,
                                            gconstpointer   buf,
@@ -181,10 +191,12 @@ static void          write_datablock_luni (GOutputStream  *output,
 
 
 static void          write_pixel_data     (GOutputStream  *output,
+                                           GimpImage      *image,
                                            GimpDrawable   *drawable,
                                            goffset        *ChanLenPosition,
                                            goffset         rowlenOffset,
-                                           gboolean        write_mask);
+                                           gboolean        write_mask,
+                                           gboolean        export_cmyk);
 
 static GimpLayer   * create_merged_image  (GimpImage      *image);
 
@@ -528,8 +540,11 @@ reshuffle_cmap_write (guchar *mapGimp)
 static void
 save_header (GOutputStream  *output,
              GimpImage      *image,
+             gboolean        export_cmyk,
              gboolean        export_duotone)
 {
+  gint nChannels;
+
   IFDBG(1) g_debug ("Function: save_header\n"
                     "\tRows: %d\n"
                     "\tColumns: %d\n"
@@ -538,18 +553,29 @@ save_header (GOutputStream  *output,
                     PSDImageData.image_height, PSDImageData.image_width,
                     PSDImageData.baseType,     PSDImageData.nChannels);
 
+  if (export_cmyk)
+    {
+      nChannels =
+        gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)) ?
+        5 : 4;
+    }
+  else
+    {
+      nChannels = (PSDImageData.nChannels + nChansLayer (PSDImageData.baseType,
+        gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)), 0));
+    }
+
   xfwrite (output, "8BPS", 4, "signature");
   write_gint16 (output, 1, "version");
   write_gint32 (output, 0, "reserved 1");      /* 6 for the 'reserved' field + 4 bytes for a long */
   write_gint16 (output, 0, "reserved 1");      /* and 2 bytes for a short */
-  write_gint16 (output, (PSDImageData.nChannels +
-                        nChansLayer (PSDImageData.baseType,
-                        gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)), 0)),
-                        "channels");
+  write_gint16 (output, nChannels, "channels");
   write_gint32 (output, PSDImageData.image_height, "rows");
   write_gint32 (output, PSDImageData.image_width, "columns");
   write_gint16 (output, 8 * get_bpc (image), "depth");
-  if (export_duotone)
+  if (export_cmyk)
+    write_gint16 (output, PSD_CMYK, "mode");
+  else if (export_duotone)
     write_gint16 (output, PSD_DUOTONE, "mode");
   else
     write_gint16 (output, gimpBaseTypeToPsdMode (PSDImageData.baseType), "mode");
@@ -634,6 +660,7 @@ save_color_mode_data (GOutputStream  *output,
 static void
 save_resources (GOutputStream  *output,
                 GimpImage      *image,
+                gboolean        export_cmyk,
                 gboolean        export_duotone)
 {
   GList        *iter;
@@ -668,124 +695,125 @@ save_resources (GOutputStream  *output,
 
 
   /* --------------- Write Channel names --------------- */
-
-  if (PSDImageData.nChannels > 0 ||
-      gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
+  if (! export_cmyk)
     {
-      xfwrite (output, "8BIM", 4, "imageresources signature");
-      write_gint16 (output, 0x03EE, "0x03EE Id"); /* 1006 */
-      /* write_pascalstring (output, Name, "Id name"); */
-      write_gint16 (output, 0, "Id name"); /* Set to null string (two zeros) */
-
-      /* Mark current position in the file */
-
-      name_sec = g_seekable_tell (G_SEEKABLE (output));
-      write_gint32 (output, 0, "0x03EE resource size");
-
-      /* Write all strings */
-
-      /* if the merged_image contains transparency, write a name for it first */
-      if (gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
-        write_string (output, "Transparency", "channel name");
-
-      for (iter = PSDImageData.lChannels; iter; iter = g_list_next (iter))
+      if (PSDImageData.nChannels > 0 ||
+          gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
         {
-          char *chName = gimp_item_get_name (iter->data);
-          write_string (output, chName, "channel name");
-          g_free (chName);
+          xfwrite (output, "8BIM", 4, "imageresources signature");
+          write_gint16 (output, 0x03EE, "0x03EE Id"); /* 1006 */
+          /* write_pascalstring (output, Name, "Id name"); */
+          write_gint16 (output, 0, "Id name"); /* Set to null string (two zeros) */
+
+          /* Mark current position in the file */
+
+          name_sec = g_seekable_tell (G_SEEKABLE (output));
+          write_gint32 (output, 0, "0x03EE resource size");
+
+          /* Write all strings */
+
+          /* if the merged_image contains transparency, write a name for it first */
+          if (gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
+            write_string (output, "Transparency", "channel name");
+
+          for (iter = PSDImageData.lChannels; iter; iter = g_list_next (iter))
+            {
+              char *chName = gimp_item_get_name (iter->data);
+              write_string (output, chName, "channel name");
+              g_free (chName);
+            }
+          /* Calculate and write actual resource's length */
+
+          eof_pos = g_seekable_tell (G_SEEKABLE (output));
+
+          g_seekable_seek (G_SEEKABLE (output),
+                           name_sec, G_SEEK_SET,
+                           NULL, NULL /*FIXME: error*/);
+          write_gint32 (output, eof_pos - name_sec - sizeof (gint32), "0x03EE resource size");
+          IFDBG(1) g_debug ("\tTotal length of 0x03EE resource: %d",
+                            (int) (eof_pos - name_sec - sizeof (gint32)));
+
+          /* Return to EOF to continue writing */
+
+          g_seekable_seek (G_SEEKABLE (output),
+                           eof_pos, G_SEEK_SET,
+                           NULL, NULL /*FIXME: error*/);
+
+          /* Pad if length is odd */
+
+          if ((eof_pos - name_sec - sizeof (gint32)) & 1)
+            write_gchar (output, 0, "pad byte");
         }
-      /* Calculate and write actual resource's length */
 
-      eof_pos = g_seekable_tell (G_SEEKABLE (output));
+      /* --------------- Write Channel properties --------------- */
 
-      g_seekable_seek (G_SEEKABLE (output),
-                       name_sec, G_SEEK_SET,
-                       NULL, NULL /*FIXME: error*/);
-      write_gint32 (output, eof_pos - name_sec - sizeof (gint32), "0x03EE resource size");
-      IFDBG(1) g_debug ("\tTotal length of 0x03EE resource: %d",
-                        (int) (eof_pos - name_sec - sizeof (gint32)));
+      if (PSDImageData.nChannels > 0 ||
+          gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
+        {
+          xfwrite (output, "8BIM", 4, "imageresources signature");
+          write_gint16 (output, 0x0435, "0x0435 Id"); /* 1077 */
+          /* write_pascalstring (output, Name, "Id name"); */
+          write_gint16 (output, 0, "Id name"); /* Set to null string (two zeros) */
+          write_gint32 (output,
+                        4  +
+                        13 * (gimp_drawable_has_alpha (
+                                GIMP_DRAWABLE (PSDImageData.merged_layer)) +
+                              PSDImageData.nChannels),
+                        "0x0435 resource size");
 
-      /* Return to EOF to continue writing */
+          /* The function of the first 4 bytes is unclear. As per
+           * load_resource_1077() in psd-image-res-load.c, it seems to be a version
+           * number that is always one.
+           */
+          write_gint32 (output, 1, "0x0435 version");
 
-      g_seekable_seek (G_SEEKABLE (output),
-                       eof_pos, G_SEEK_SET,
-                       NULL, NULL /*FIXME: error*/);
+          /* Write all channel properties */
 
-      /* Pad if length is odd */
+          #define DOUBLE_TO_INT16(x) ROUND (SAFE_CLAMP (x, 0.0, 1.0) * 0xffff)
 
-      if ((eof_pos - name_sec - sizeof (gint32)) & 1)
-        write_gchar (output, 0, "pad byte");
+          /* if the merged_image contains transparency, write its properties first */
+          if (gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
+            {
+              write_gint16 (output, PSD_CS_RGB, "channel color space");
+              write_gint16 (output, DOUBLE_TO_INT16 (1.0), "channel color r");
+              write_gint16 (output, DOUBLE_TO_INT16 (0.0), "channel color g");
+              write_gint16 (output, DOUBLE_TO_INT16 (0.0), "channel color b");
+              write_gint16 (output, 0,                     "channel color padding");
+              write_gint16 (output, 100,                   "channel opacity");
+              write_gchar  (output, 1,                     "channel mode");
+            }
+
+          for (iter = PSDImageData.lChannels; iter; iter = g_list_next (iter))
+            {
+              GimpChannel *channel = iter->data;
+              GimpRGB      color;
+              gdouble      opacity;
+
+              gimp_channel_get_color (channel, &color);
+              opacity = gimp_channel_get_opacity (channel);
+
+              write_gint16 (output, PSD_CS_RGB,                "channel color space");
+              write_gint16 (output, DOUBLE_TO_INT16 (color.r), "channel color r");
+              write_gint16 (output, DOUBLE_TO_INT16 (color.g), "channel color g");
+              write_gint16 (output, DOUBLE_TO_INT16 (color.b), "channel color b");
+              write_gint16 (output, 0,                         "channel color padding");
+              write_gint16 (output, ROUND (opacity),           "channel opacity");
+              write_gchar  (output, 1,                         "channel mode");
+            }
+
+          #undef DOUBLE_TO_INT16
+
+          /* Pad if length is odd */
+
+          if (g_seekable_tell (G_SEEKABLE (output)) & 1)
+            write_gchar (output, 0, "pad byte");
+        }
     }
-
-  /* --------------- Write Channel properties --------------- */
-
-  if (PSDImageData.nChannels > 0 ||
-      gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
-    {
-      xfwrite (output, "8BIM", 4, "imageresources signature");
-      write_gint16 (output, 0x0435, "0x0435 Id"); /* 1077 */
-      /* write_pascalstring (output, Name, "Id name"); */
-      write_gint16 (output, 0, "Id name"); /* Set to null string (two zeros) */
-      write_gint32 (output,
-                    4  +
-                    13 * (gimp_drawable_has_alpha (
-                            GIMP_DRAWABLE (PSDImageData.merged_layer)) +
-                          PSDImageData.nChannels),
-                    "0x0435 resource size");
-
-      /* The function of the first 4 bytes is unclear. As per
-       * load_resource_1077() in psd-image-res-load.c, it seems to be a version
-       * number that is always one.
-       */
-      write_gint32 (output, 1, "0x0435 version");
-
-      /* Write all channel properties */
-
-      #define DOUBLE_TO_INT16(x) ROUND (SAFE_CLAMP (x, 0.0, 1.0) * 0xffff)
-
-      /* if the merged_image contains transparency, write its properties first */
-      if (gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)))
-        {
-          write_gint16 (output, PSD_CS_RGB, "channel color space");
-          write_gint16 (output, DOUBLE_TO_INT16 (1.0), "channel color r");
-          write_gint16 (output, DOUBLE_TO_INT16 (0.0), "channel color g");
-          write_gint16 (output, DOUBLE_TO_INT16 (0.0), "channel color b");
-          write_gint16 (output, 0,                     "channel color padding");
-          write_gint16 (output, 100,                   "channel opacity");
-          write_gchar  (output, 1,                     "channel mode");
-        }
-
-      for (iter = PSDImageData.lChannels; iter; iter = g_list_next (iter))
-        {
-          GimpChannel *channel = iter->data;
-          GimpRGB      color;
-          gdouble      opacity;
-
-          gimp_channel_get_color (channel, &color);
-          opacity = gimp_channel_get_opacity (channel);
-
-          write_gint16 (output, PSD_CS_RGB,                "channel color space");
-          write_gint16 (output, DOUBLE_TO_INT16 (color.r), "channel color r");
-          write_gint16 (output, DOUBLE_TO_INT16 (color.g), "channel color g");
-          write_gint16 (output, DOUBLE_TO_INT16 (color.b), "channel color b");
-          write_gint16 (output, 0,                         "channel color padding");
-          write_gint16 (output, ROUND (opacity),           "channel opacity");
-          write_gchar  (output, 1,                         "channel mode");
-        }
-
-      #undef DOUBLE_TO_INT16
-
-      /* Pad if length is odd */
-
-      if (g_seekable_tell (G_SEEKABLE (output)) & 1)
-        write_gchar (output, 0, "pad byte");
-    }
-
   /* --------------- Write Guides --------------- */
   if (gimp_image_find_next_guide (image, 0))
     {
       gint n_guides = 0;
-      gint guide_id =0;
+      gint guide_id = 0;
 
       /* Count the guides */
       while ((guide_id = gimp_image_find_next_guide(image, guide_id)))
@@ -822,23 +850,21 @@ save_resources (GOutputStream  *output,
                         (int) sizeof (gint16));
     }
 
+  /* --------------- Write paths ------------------- */
+  save_paths (output, image);
+
   /* --------------- Write resolution data ------------------- */
   {
-    gdouble xres = 0, yres = 0;
-    guint32 xres_fix, yres_fix;
+    gdouble  xres = 0, yres = 0;
+    guint32  xres_fix, yres_fix;
     GimpUnit g_unit;
-    gint16 psd_unit;
+    gint16   psd_unit;
 
     g_unit = gimp_image_get_unit (image);
     gimp_image_get_resolution (image, &xres, &yres);
 
     if (g_unit == GIMP_UNIT_MM)
       {
-        gdouble factor = gimp_unit_get_factor (g_unit) / 10.0;
-
-        xres /= factor;
-        yres /= factor;
-
         psd_unit = PSD_UNIT_CM;
       }
     else
@@ -846,6 +872,8 @@ save_resources (GOutputStream  *output,
         psd_unit = PSD_UNIT_INCH;
       }
 
+    /* Don't convert resolution based on g_unit which is a display unit.
+     * PSD resolution is always in pixels/inch. */
     xres_fix = xres * 65536.0 + .5; /* Convert to 16.16 fixed point */
     yres_fix = yres * 65536.0 + .5; /* Convert to 16.16 fixed point */
 
@@ -853,10 +881,10 @@ save_resources (GOutputStream  *output,
     write_gint16(output, 0x03ed, "0x03ed Id (resolution)"); /* 1005 */
     write_gint16 (output, 0, "Id name"); /* Set to null string (two zeros) */
     write_gint32 (output, 16, "0x0400 resource size");
-    write_gint32 (output,  xres_fix, "hRes (16.16 fixed point)");
+    write_gint32 (output, xres_fix, "hRes (16.16 fixed point)");
     write_gint16 (output, psd_unit, "hRes unit");
     write_gint16 (output, psd_unit, "width unit");
-    write_gint32 (output,  yres_fix, "vRes (16.16 fixed point)");
+    write_gint32 (output, yres_fix, "vRes (16.16 fixed point)");
     write_gint16 (output, psd_unit, "vRes unit");
     write_gint16 (output, psd_unit, "height unit");
   }
@@ -917,6 +945,14 @@ save_resources (GOutputStream  *output,
     if (! export_duotone)
       profile = gimp_image_get_effective_color_profile (image);
 
+    if (export_cmyk)
+      {
+        profile = gimp_image_get_simulation_profile (image);
+
+        if (! gimp_color_profile_is_cmyk (profile))
+          g_object_unref (profile);
+      }
+
     if (profile)
       {
         const guint8 *icc_data;
@@ -1039,9 +1075,173 @@ get_compress_channel_data (guchar  *channel_data,
   return len;
 }
 
+/* Ported /from plug-ins/file-tiff/file-tiff-save.c */
+static void
+double_to_psd_fixed (gdouble  value,
+                     gchar   *target)
+{
+  gdouble in, frac;
+  gint    i, f;
+
+  frac = modf (value, &in);
+  if (frac < 0)
+    {
+      in -= 1;
+      frac += 1;
+    }
+
+  i = (gint) CLAMP (in, -16, 15);
+  f = CLAMP ((gint) (frac * 0xFFFFFF), 0, 0xFFFFFF);
+
+  target[0] = i & 0xFF;
+  target[1] = (f >> 16) & 0xFF;
+  target[2] = (f >>  8) & 0xFF;
+  target[3] = f & 0xFF;
+}
+
+/* Ported from /plug-ins/file-tiff/file-tiff-save.c */
+static void
+save_paths (GOutputStream  *output,
+            GimpImage      *image)
+{
+  gshort  id     = 0x07D0; /* Photoshop paths have IDs >= 2000 */
+  gdouble width  = gimp_image_get_width (image);
+  gdouble height = gimp_image_get_height (image);
+  GList  *vectors;
+  GList  *iter;
+  gint    v;
+  gint    num_strokes;
+  gint   *strokes;
+  gint    s;
+
+  vectors = gimp_image_list_vectors (image);
+
+  if (! vectors)
+    return;
+
+  /* Only up to 997 paths supported */
+  for (iter = vectors, v = 0;
+       iter && v <= 997;
+       iter = g_list_next (iter), v++)
+    {
+      GString *data;
+      gchar   *name, *nameend;
+      gsize    len;
+      gint     lenpos;
+      gchar    pointrecord[26] = { 0, };
+      gchar   *tmpname;
+      GError  *err = NULL;
+
+      data = g_string_new ("8BIM");
+      g_string_append_c (data, id / 256);
+      g_string_append_c (data, id % 256);
+
+      /*
+       * - use iso8859-1 if possible
+       * - otherwise use UTF-8, prepended with \xef\xbb\xbf (Byte-Order-Mark)
+       */
+      name = gimp_item_get_name (iter->data);
+      tmpname = g_convert (name, -1, "iso8859-1", "utf-8", NULL, &len, &err);
+
+      if (tmpname && err == NULL)
+        {
+          g_string_append_c (data, MIN (len, 255));
+          g_string_append_len (data, tmpname, MIN (len, 255));
+          g_free (tmpname);
+        }
+      else
+        {
+          /* conversion failed, we fall back to UTF-8 */
+          len = g_utf8_strlen (name, 255 - 3);  /* need three marker-bytes */
+
+          nameend = g_utf8_offset_to_pointer (name, len);
+          len = nameend - name; /* in bytes */
+          g_assert (len + 3 <= 255);
+
+          g_string_append_c (data, len + 3);
+          g_string_append_len (data, "\xEF\xBB\xBF", 3); /* Unicode 0xfeff */
+          g_string_append_len (data, name, len);
+
+          if (tmpname)
+            g_free (tmpname);
+        }
+
+      if (data->len % 2)  /* padding to even size */
+        g_string_append_c (data, 0);
+      g_free (name);
+
+      lenpos = data->len;
+      g_string_append_len (data, "\0\0\0\0", 4); /* will be filled in later */
+      len = data->len; /* to calculate the data size later */
+
+      pointrecord[1] = 6;  /* fill rule record */
+      g_string_append_len (data, pointrecord, 26);
+
+      strokes = gimp_vectors_get_strokes (iter->data, &num_strokes);
+
+      for (s = 0; s < num_strokes; s++)
+        {
+          GimpVectorsStrokeType type;
+          gdouble  *points;
+          gint      num_points;
+          gboolean  closed;
+          gint      p = 0;
+
+          type = gimp_vectors_stroke_get_points (iter->data, strokes[s],
+                                                 &num_points, &points, &closed);
+
+          if (type != GIMP_VECTORS_STROKE_TYPE_BEZIER ||
+              num_points > 65535 ||
+              num_points % 6)
+            {
+              g_printerr ("psd-save: unsupported stroke type: "
+                          "%d (%d points)\n", type, num_points);
+              continue;
+            }
+
+          memset (pointrecord, 0, 26);
+          pointrecord[1] = closed ? 0 : 3;
+          pointrecord[2] = (num_points / 6) / 256;
+          pointrecord[3] = (num_points / 6) % 256;
+          g_string_append_len (data, pointrecord, 26);
+
+          for (p = 0; p < num_points; p += 6)
+            {
+              pointrecord[1] = closed ? 2 : 5;
+
+              double_to_psd_fixed (points[p+1] / height, pointrecord + 2);
+              double_to_psd_fixed (points[p+0] / width,  pointrecord + 6);
+              double_to_psd_fixed (points[p+3] / height, pointrecord + 10);
+              double_to_psd_fixed (points[p+2] / width,  pointrecord + 14);
+              double_to_psd_fixed (points[p+5] / height, pointrecord + 18);
+              double_to_psd_fixed (points[p+4] / width,  pointrecord + 22);
+
+              g_string_append_len (data, pointrecord, 26);
+            }
+        }
+
+      g_free (strokes);
+
+      /* fix up the length */
+
+      len = data->len - len;
+      data->str[lenpos + 0] = (len & 0xFF000000) >> 24;
+      data->str[lenpos + 1] = (len & 0x00FF0000) >> 16;
+      data->str[lenpos + 2] = (len & 0x0000FF00) >>  8;
+      data->str[lenpos + 3] = (len & 0x000000FF) >>  0;
+
+      xfwrite (output, data->str, data->len, "path resources data");
+      g_string_free (data, TRUE);
+      id += 0x01;
+    }
+
+  g_list_free (vectors);
+}
+
 static void
 save_layer_and_mask (GOutputStream  *output,
-                     GimpImage      *image)
+                     GimpImage      *image,
+                     gboolean        export_cmyk)
 {
   gint           i,j;
   gint           idChannel;
@@ -1155,6 +1355,15 @@ save_layer_and_mask (GOutputStream  *output,
                                     gimp_drawable_has_alpha (GIMP_DRAWABLE (psd_layer->layer)),
                                     hasMask);
 
+      /* Manually set channels to 4 or 5 when export as CMYK;
+       * Can be removed once CMYK channels are accessible in GIMP
+       */
+      if (export_cmyk)
+        {
+          nChannelsLayer =
+            gimp_drawable_has_alpha (GIMP_DRAWABLE (psd_layer->layer)) ?
+            5 : 4;
+        }
 
       write_gint16 (output, nChannelsLayer, "Number channels in the layer");
       IFDBG(1) g_debug ("\t\tNumber of channels: %d", nChannelsLayer);
@@ -1355,8 +1564,8 @@ save_layer_and_mask (GOutputStream  *output,
       gimp_progress_update ((PSDImageData.nLayers - i - 1.0) / (PSDImageData.nLayers + 1.0));
 
       IFDBG(1) g_debug ("\t\tWriting pixel data for layer slot %d", i-1);
-      write_pixel_data (output, GIMP_DRAWABLE (psd_layer->layer), ChannelLengthPos[i-1], 0,
-                        psd_layer->type != PSD_LAYER_TYPE_GROUP_END);
+      write_pixel_data (output, image, GIMP_DRAWABLE (psd_layer->layer), ChannelLengthPos[i-1], 0,
+                        psd_layer->type != PSD_LAYER_TYPE_GROUP_END, export_cmyk);
       g_free (ChannelLengthPos[i-1]);
     }
 
@@ -1390,28 +1599,33 @@ save_layer_and_mask (GOutputStream  *output,
 
 static void
 write_pixel_data (GOutputStream  *output,
+                  GimpImage      *image,
                   GimpDrawable   *drawable,
                   goffset        *ChanLenPosition,
                   goffset         ltable_offset,
-                  gboolean        write_mask)
+                  gboolean        write_mask,
+                  gboolean        export_cmyk)
 {
-  GeglBuffer    *buffer = gimp_drawable_get_buffer (drawable);
-  const Babl    *format;
-  GimpLayerMask *mask;
-  gint32         tile_height = gimp_tile_height ();
-  gint32         height = gegl_buffer_get_height (buffer);
-  gint32         width  = gegl_buffer_get_width (buffer);
-  gint32         bytes;
-  gint32         components;
-  gint32         bpc;
-  gint32         colors;
-  gint32         y;
-  gsize          len;                  /* Length of compressed data */
-  gint16        *LengthsTable;         /* Lengths of every compressed row */
-  guchar        *rledata;              /* Compressed data from a region */
-  guchar        *data;                 /* Temporary copy of pixel data */
-  goffset        length_table_pos;     /* position in file of the length table */
-  int            i, j;
+  GeglBuffer       *buffer = gimp_drawable_get_buffer (drawable);
+  const Babl       *format;
+  const Babl       *space  = NULL;
+  const Babl       *type;
+  GimpColorProfile *profile;
+  GimpLayerMask    *mask;
+  gint32            tile_height = gimp_tile_height ();
+  gint32            height = gegl_buffer_get_height (buffer);
+  gint32            width  = gegl_buffer_get_width (buffer);
+  gint32            bytes;
+  gint32            components;
+  gint32            bpc;
+  gint32            colors;
+  gint32            y;
+  gsize             len;                  /* Length of compressed data */
+  gint16           *LengthsTable;         /* Lengths of every compressed row */
+  guchar           *rledata;              /* Compressed data from a region */
+  guchar           *data;                 /* Temporary copy of pixel data */
+  goffset           length_table_pos;     /* position in file of the length table */
+  int               i, j;
 
   IFDBG(1) g_debug ("Function: write_pixel_data, drw %d, lto %" G_GOFFSET_FORMAT,
                     gimp_item_get_id (GIMP_ITEM (drawable)), ltable_offset);
@@ -1433,6 +1647,43 @@ write_pixel_data (GOutputStream  *output,
   else
     format = get_pixel_format (drawable);
 
+  if (export_cmyk && ! gimp_item_is_channel (GIMP_ITEM (drawable)))
+    {
+      profile = gimp_image_get_simulation_profile (image);
+      if (profile && gimp_color_profile_is_cmyk (profile))
+        space = gimp_color_profile_get_space (profile,
+                                              gimp_image_get_simulation_intent (image),
+                                              NULL);
+      if (profile)
+        g_object_unref (profile);
+
+      if (get_bpc (image) == 1)
+        type = babl_type ("u8");
+      else
+        type = babl_type ("u16");
+
+      if (gimp_drawable_has_alpha (drawable))
+        format = babl_format_new (babl_model ("cmykA"),
+                                  type,
+                                  babl_component ("cyan"),
+                                  babl_component ("magenta"),
+                                  babl_component ("yellow"),
+                                  babl_component ("key"),
+                                  babl_component ("A"),
+                                  NULL);
+      else
+        format = babl_format_new (babl_model ("cmyk"),
+                                  type,
+                                  babl_component ("cyan"),
+                                  babl_component ("magenta"),
+                                  babl_component ("yellow"),
+                                  babl_component ("key"),
+                                  NULL);
+
+      format = babl_format_with_space (babl_format_get_encoding (format),
+                                       space);
+    }
+
   bytes      = babl_format_get_bytes_per_pixel (format);
   components = babl_format_get_n_components    (format);
   bpc        = bytes / components;
@@ -1638,7 +1889,8 @@ write_pixel_data (GOutputStream  *output,
 
 static void
 save_data (GOutputStream  *output,
-           GimpImage      *image)
+           GimpImage      *image,
+           gboolean        export_cmyk)
 {
   GList  *iter;
   gint    ChanCount;
@@ -1667,8 +1919,8 @@ save_data (GOutputStream  *output,
       write_gint16 (output, 0, "junk line lengths");
 
   IFDBG(1) g_debug ("\t\tWriting compressed image data");
-  write_pixel_data (output, GIMP_DRAWABLE (PSDImageData.merged_layer),
-                    NULL, offset, FALSE);
+  write_pixel_data (output, image, GIMP_DRAWABLE (PSDImageData.merged_layer),
+                    NULL, offset, FALSE, export_cmyk);
 
   chan = nChansLayer (PSDImageData.baseType,
                       gimp_drawable_has_alpha (GIMP_DRAWABLE (PSDImageData.merged_layer)), 0);
@@ -1676,8 +1928,9 @@ save_data (GOutputStream  *output,
   for (iter = PSDImageData.lChannels; iter; iter = g_list_next (iter))
     {
       IFDBG(1) g_debug ("\t\tWriting compressed channel data for channel %d", i);
-      write_pixel_data (output, iter->data, NULL,
-                        offset + 2*imageHeight*chan, FALSE); //check how imgs are channels here
+      write_pixel_data (output, image, iter->data, NULL,
+                        offset + 2*imageHeight*chan, FALSE,
+                        export_cmyk); //check how imgs are channels here
       chan++;
     }
 }
@@ -1789,15 +2042,20 @@ save_image (GFile      *file,
   GeglBuffer     *buffer;
   GList          *iter;
   GError         *local_error = NULL;
-  GimpParasite   *parasite = NULL;
+  GimpParasite   *parasite    = NULL;
+  gboolean        config_cmyk;
   gboolean        config_duotone;
 
   g_object_get (config,
+                "cmyk",    &config_cmyk,
                 "duotone", &config_duotone,
                 NULL);
 
   IFDBG(1) g_debug ("Function: save_image");
 
+  if (config_cmyk)
+    config_duotone = FALSE;
+
   if (gimp_image_get_width (image) > 30000 ||
       gimp_image_get_height (image) > 30000)
     {
@@ -1879,20 +2137,20 @@ save_image (GFile      *file,
   IFDBG(1) g_debug ("\tFile '%s' has been opened",
                     gimp_file_get_utf8_name (file));
 
-  save_header (output, image, config_duotone);
+  save_header (output, image, config_cmyk, config_duotone);
   save_color_mode_data (output, image, config_duotone);
-  save_resources (output, image, config_duotone);
+  save_resources (output, image, config_cmyk, config_duotone);
 
   /* PSD format does not support layers in indexed images */
 
   if (PSDImageData.baseType == GIMP_INDEXED)
     write_gint32 (output, 0, "layers info section length");
   else
-    save_layer_and_mask (output, image);
+    save_layer_and_mask (output, image, config_cmyk);
 
   /* If this is an indexed image, write now channel and layer info */
 
-  save_data (output, image);
+  save_data (output, image, config_cmyk);
 
   /* Delete merged image now */
 
@@ -2082,33 +2340,97 @@ save_dialog (GimpImage     *image,
              GimpProcedure *procedure,
              GObject       *config)
 {
-  GtkWidget *dialog;
-  GtkWidget *duotone_notice;
-  gboolean   run;
+  GtkWidget        *dialog;
+  GtkWidget        *duotone_notice;
+  GtkWidget        *profile_label;
+  GimpColorProfile *cmyk_profile;
+  GimpParasite     *parasite         = NULL;
+  gboolean          has_duotone_data = FALSE;
+  gboolean          run;
 
   dialog = gimp_procedure_dialog_new (procedure,
                                       GIMP_PROCEDURE_CONFIG (config),
                                       _("Export Image as PSD"));
 
-  /* Profile label */
-  duotone_notice = gimp_procedure_dialog_get_label (GIMP_PROCEDURE_DIALOG (dialog),
-                                                    "duotone-notice",
-                                                    _("Duotone color space information "
-                                                    "from the original\nimported image "
-                                                    "will be used."));
-  gtk_label_set_xalign (GTK_LABEL (duotone_notice), 0.0);
-  gtk_label_set_ellipsize (GTK_LABEL (duotone_notice), PANGO_ELLIPSIZE_END);
-  gimp_label_set_attributes (GTK_LABEL (duotone_notice),
+  /* CMYK profile label. */
+  profile_label = gimp_procedure_dialog_get_label (GIMP_PROCEDURE_DIALOG (dialog),
+                                                   "profile-label", _("No soft-proofing profile"));
+  gtk_label_set_xalign (GTK_LABEL (profile_label), 0.0);
+  gtk_label_set_ellipsize (GTK_LABEL (profile_label), PANGO_ELLIPSIZE_END);
+  gimp_label_set_attributes (GTK_LABEL (profile_label),
                              PANGO_ATTR_STYLE, PANGO_STYLE_ITALIC,
                              -1);
-
+  gimp_help_set_help_data (profile_label,
+                           _("Name of the color profile used for CMYK export."), NULL);
   gimp_procedure_dialog_fill_frame (GIMP_PROCEDURE_DIALOG (dialog),
-                                    "duotone-frame", "duotone", FALSE,
-                                    "duotone-notice");
+                                    "cmyk-frame", "cmyk", FALSE,
+                                    "profile-label");
+  cmyk_profile = gimp_image_get_simulation_profile (image);
+  if (cmyk_profile)
+    {
+      if (gimp_color_profile_is_cmyk (cmyk_profile))
+        {
+          gchar *label_text;
 
-  gimp_procedure_dialog_fill (GIMP_PROCEDURE_DIALOG (dialog),
-                              "duotone-frame",
-                              NULL);
+          label_text = g_strdup_printf (_("Profile: %s"),
+                                        gimp_color_profile_get_label (cmyk_profile));
+          gtk_label_set_text (GTK_LABEL (profile_label), label_text);
+          gimp_label_set_attributes (GTK_LABEL (profile_label),
+                                     PANGO_ATTR_STYLE, PANGO_STYLE_NORMAL,
+                                     -1);
+          g_free (label_text);
+        }
+      g_object_unref (cmyk_profile);
+    }
+
+  /* Only show dialog if image is grayscale and duotone color space
+   * information was attached from the original image imported
+   */
+   parasite = gimp_image_get_parasite (image, PSD_PARASITE_DUOTONE_DATA);
+   if (parasite)
+     {
+       if (gimp_image_get_base_type (image) == GIMP_GRAY)
+         {
+           has_duotone_data = TRUE;
+
+           /* Duotone Option label */
+           duotone_notice = gimp_procedure_dialog_get_label (GIMP_PROCEDURE_DIALOG (dialog),
+                                                             "duotone-notice",
+                                                             _("Duotone color space information "
+                                                             "from the original\nimported image "
+                                                             "will be used."));
+           gtk_label_set_xalign (GTK_LABEL (duotone_notice), 0.0);
+           gtk_label_set_ellipsize (GTK_LABEL (duotone_notice), PANGO_ELLIPSIZE_END);
+           gimp_label_set_attributes (GTK_LABEL (duotone_notice),
+                                      PANGO_ATTR_STYLE, PANGO_STYLE_ITALIC,
+                                      -1);
+
+           gimp_procedure_dialog_fill_frame (GIMP_PROCEDURE_DIALOG (dialog),
+                                             "duotone-frame", "duotone", FALSE,
+                                             "duotone-notice");
+
+           /* Prevent you from setting both Duotone and CMYK exports */
+           gimp_procedure_dialog_set_sensitive (GIMP_PROCEDURE_DIALOG (dialog),
+                                                "duotone",
+                                                TRUE, config, "cmyk", TRUE);
+           gimp_procedure_dialog_set_sensitive (GIMP_PROCEDURE_DIALOG (dialog),
+                                                "cmyk",
+                                                TRUE, config, "duotone", TRUE);
+
+         }
+
+       gimp_parasite_free (parasite);
+     }
+
+  if (has_duotone_data)
+    gimp_procedure_dialog_fill (GIMP_PROCEDURE_DIALOG (dialog),
+                                "cmyk-frame",
+                                "duotone-frame",
+                                NULL);
+  else
+    gimp_procedure_dialog_fill (GIMP_PROCEDURE_DIALOG (dialog),
+                                "cmyk-frame",
+                                NULL);
 
   gtk_widget_show (dialog);
 
@@ -2117,4 +2439,4 @@ save_dialog (GimpImage     *image,
   gtk_widget_destroy (dialog);
 
   return run;
-}
+}
\ No newline at end of file
